# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "Address",
    "AddressReader",
    "AddressRow",
    "AddressWriter",
    "Compressor",
    "Decompressor",
    "InvalidQuantumGraphFileError",
    "MultiblockReader",
    "MultiblockWriter",
)

import dataclasses
import logging
import tempfile
import uuid
import zipfile
from collections.abc import Iterator, Set
from contextlib import contextmanager
from io import BufferedReader, BytesIO
from operator import attrgetter
from typing import IO, Protocol, TypeVar

import pydantic

_LOG = logging.getLogger(__name__)


_T = TypeVar("_T", bound=pydantic.BaseModel)


type UUID_int = int

MAX_UUID_INT: UUID_int = 2**128


DEFAULT_PAGE_SIZE: int = 5_000_000
"""Default page size for reading chunks of quantum graph files.

This is intended to be large enough to avoid any possibility of individual
reads suffering from per-seek overheads, especially in network file access,
while still being small enough to only minimally slow down tiny reads of
individual quanta (especially for execution).
"""


class Compressor(Protocol):
    """A protocol for objects with a ``compress`` method that takes and returns
    `bytes`.
    """

    def compress(self, data: bytes) -> bytes:
        """Compress the given data.

        Parameters
        ----------
        data : `bytes`
            Uncompressed data.

        Returns
        -------
        compressed : `bytes`
            Compressed data.
        """
        ...


class Decompressor(Protocol):
    """A protocol for objects with a `decompress` method that takes and returns
    `bytes`.
    """

    def decompress(self, data: bytes) -> bytes:
        """Decompress the given data.

        Parameters
        ----------
        data : `bytes`
            Compressed data.

        Returns
        -------
        decompressed : `bytes`
            Uncompressed data.
        """
        ...


class InvalidQuantumGraphFileError(RuntimeError):
    """An exception raised when a quantum graph file has internal
    inconsistencies or does not actually appear to be a quantum graph file.
    """


@dataclasses.dataclass(slots=True)
class Address:
    """Struct that holds an address into a multi-block file."""

    offset: int = 0
    """Byte offset for the block."""

    size: int = 0
    """Size of the block.

    This always includes the size of the tiny header that records the block
    size.  That header does not include the size of the header, so these sizes
    differ by the ``int_size`` used to write the multi-block file.

    A size of zero is used (with, by convention, an offset of zero) to indicate
    an absent block.
    """

    def __str__(self) -> str:
        return f"{self.offset:06}[{self.size:06}]"


@dataclasses.dataclass(slots=True)
class AddressRow:
    """The in-memory representation of a single row in an address file."""

    key: uuid.UUID
    """Universally unique identifier for this row."""

    index: int
    """Monotonically increasing integer ID; unique within this file only."""

    addresses: list[Address] = dataclasses.field(default_factory=list)
    """Offsets and sizes into multi-block files."""

    def write(self, stream: IO[bytes], int_size: int) -> None:
        """Write this address row to a file-like object.

        Parameters
        ----------
        stream : `typing.IO` [ `bytes` ]
            Binary file-like object.
        int_size : `int`
            Number of bytes to use for all integers.
        """
        stream.write(self.key.bytes)
        stream.write(self.index.to_bytes(int_size))
        for address in self.addresses:
            stream.write(address.offset.to_bytes(int_size))
            stream.write(address.size.to_bytes(int_size))

    @classmethod
    def read(cls, stream: IO[bytes], n_addresses: int, int_size: int) -> AddressRow:
        """Read this address row from a file-like object.

        Parameters
        ----------
        stream : `typing.IO` [ `bytes` ]
            Binary file-like object.
        n_addresses : `int`
            Number of addresses included in each row.
        int_size : `int`
            Number of bytes to use for all integers.
        """
        key = uuid.UUID(int=int.from_bytes(stream.read(16)))
        index = int.from_bytes(stream.read(int_size))
        row = AddressRow(key, index)
        for _ in range(n_addresses):
            offset = int.from_bytes(stream.read(int_size))
            size = int.from_bytes(stream.read(int_size))
            row.addresses.append(Address(offset, size))
        return row

    def __str__(self) -> str:
        return f"{self.key} {self.index:06} {' '.join(str(a) for a in self.addresses)}"


@dataclasses.dataclass
class AddressWriter:
    """A helper object for writing address files for multi-block files."""

    addresses: list[dict[uuid.UUID, Address]] = dataclasses.field(default_factory=list)
    """Addresses to store with each UUID.

    Every key in one of these dictionaries must have an entry in ``indices``.
    The converse is not true.
    """

    def write(self, stream: IO[bytes], int_size: int, all_ids: Set[uuid.UUID] | None = None) -> None:
        """Write all addresses to a file-like object.

        Parameters
        ----------
        stream : `typing.IO` [ `bytes` ]
            Binary file-like object.
        int_size : `int`
            Number of bytes to use for all integers.
        all_ids : `~collections.abc.Set` [`uuid.UUID`], optional
            Set of the union of all UUIDs in any dictionary from a call to
            `get_all_ids`.
        """
        if all_ids is None:
            all_ids = self.get_all_ids()
        stream.write(int_size.to_bytes(1))
        stream.write(len(all_ids).to_bytes(int_size))
        stream.write(len(self.addresses).to_bytes(int_size))
        empty_address = Address()
        for n, key in enumerate(sorted(all_ids, key=attrgetter("int"))):
            row = AddressRow(key, n, [m.get(key, empty_address) for m in self.addresses])
            _LOG.debug("Wrote address %s.", row)
            row.write(stream, int_size)

    def write_to_zip(self, zf: zipfile.ZipFile, name: str, int_size: int) -> None:
        """Write all addresses to a file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to add the file to.
        name : `str`
            Base name for the address file; an extension will be added.
        int_size : `int`
            Number of bytes to use for all integers.
        """
        all_ids = self.get_all_ids()
        zip_info = zipfile.ZipInfo(f"{name}.addr")
        row_size = AddressReader.compute_row_size(int_size, len(self.addresses))
        zip_info.file_size = AddressReader.compute_header_size(int_size) + len(all_ids) * row_size
        with zf.open(zip_info, mode="w") as stream:
            self.write(stream, int_size=int_size, all_ids=all_ids)

    def get_all_ids(self) -> Set[uuid.UUID]:
        """Return all IDs used by any address dictionary.

        Returns
        -------
        all_ids : `~collections.abc.Set` [`uuid.UUID`]
            Set of all IDs.
        """
        all_ids: set[uuid.UUID] = set()
        for address_map in self.addresses:
            all_ids.update(address_map.keys())
        return all_ids


@dataclasses.dataclass
class AddressPage:
    """A page of addresses in the `AddressReader`."""

    file_offset: int
    """Offset in bytes to this page from the beginning of the file."""

    begin: int
    """Index of the first row in this page."""

    n_rows: int
    """Number of rows in this page."""

    read: bool = False
    """Whether this page has already been read."""

    @property
    def end(self) -> int:
        """One past the last row index in this page."""
        return self.begin + self.n_rows


@dataclasses.dataclass
class PageBounds:
    """A page index and the UUID interval that page covers."""

    page_index: int
    """Index into the page array."""

    uuid_int_begin: UUID_int
    """Integer representation of the smallest UUID in this page."""

    uuid_int_end: UUID_int
    """One larger than the integer representation of the largest UUID in this
    page.
    """

    def __str__(self) -> str:
        return f"{self.page_index} [{self.uuid_int_begin:x}:{self.uuid_int_end:x}]"


@dataclasses.dataclass
class AddressReader:
    """A helper object for reading address files for multi-block files."""

    stream: IO[bytes]
    """Stream to read from."""

    int_size: int
    """Size of each integer in bytes."""

    n_rows: int
    """Number of rows in the file."""

    n_addresses: int
    """Number of addresses in each row."""

    rows_per_page: int
    """Number of addresses in each page."""

    rows: dict[uuid.UUID, AddressRow] = dataclasses.field(default_factory=dict)
    """Rows that have already been read."""

    pages: list[AddressPage] = dataclasses.field(default_factory=list)
    """Descriptions of the file offsets and integer row indexes of pages and
    flags for whether they have been read already.
    """

    page_bounds: dict[int, PageBounds] = dataclasses.field(default_factory=dict)
    """Mapping from page index to page boundary information."""

    @classmethod
    def from_stream(
        cls, stream: IO[bytes], *, page_size: int, n_addresses: int, int_size: int
    ) -> AddressReader:
        """Construct from a stream by reading the header.

        Parameters
        ----------
        stream : `typing.IO` [ `bytes` ]
            File-like object to read from.
        page_size : `int`
            Approximate number of bytes to read at a time when searching for an
            address.
        n_addresses : `int`
            Number of addresses to expect per row.  This is checked against
            the size embedded in the file.
        int_size : `int`
            Number of bytes to use for all integers.  This is checked against
            the size embedded in the file.
        """
        header_size = cls.compute_header_size(int_size)
        row_size = cls.compute_row_size(int_size, n_addresses)
        # Read the raw header page.
        header_page_data = stream.read(header_size)
        if len(header_page_data) < header_size:
            raise InvalidQuantumGraphFileError("Address file unexpectedly truncated.")
        # Interpret the raw header data and initialize the reader instance.
        header_page_stream = BytesIO(header_page_data)
        file_int_size = int.from_bytes(header_page_stream.read(1))
        if file_int_size != int_size:
            raise InvalidQuantumGraphFileError(
                f"int size in address file ({file_int_size}) does not match int size in header ({int_size})."
            )
        n_rows = int.from_bytes(header_page_stream.read(int_size))
        file_n_addresses = int.from_bytes(header_page_stream.read(int_size))
        if file_n_addresses != n_addresses:
            raise InvalidQuantumGraphFileError(
                f"Incorrect number of addresses per row: expected {n_addresses}, got {file_n_addresses}."
            )
        rows_per_page = max(page_size // row_size, 1)
        # Construct an instance.
        self = cls(stream, int_size, n_rows, n_addresses, rows_per_page=rows_per_page)
        # Calculate positions of each page of rows.
        row_index = 0
        file_offset = header_size
        while row_index < n_rows:
            self.pages.append(AddressPage(file_offset=file_offset, begin=row_index, n_rows=rows_per_page))
            row_index += rows_per_page
            file_offset += rows_per_page * row_size
        if row_index != n_rows:
            # Last page was too big.
            self.pages[-1].n_rows -= row_index - n_rows
            assert sum(p.n_rows for p in self.pages) == n_rows, "Bad logic setting page row counts."
        return self

    @classmethod
    @contextmanager
    def open_in_zip(
        cls,
        zf: zipfile.ZipFile,
        name: str,
        *,
        page_size: int,
        n_addresses: int,
        int_size: int,
    ) -> Iterator[AddressReader]:
        """Make a reader for an address file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to read the file from.
        name : `str`
            Base name for the address file; an extension will be added.
        page_size : `int`
            Approximate number of bytes to read at a time when searching for an
            address.
        n_addresses : `int`
            Number of addresses to expect per row.  This is checked against
            the size embedded in the file.
        int_size : `int`
            Number of bytes to use for all integers.  This is checked against
            the size embedded in the file.

        Returns
        -------
        reader : `contextlib.AbstractContextManager` [ `AddressReader` ]
            Context manager that returns a reader when entered.
        """
        with zf.open(f"{name}.addr", mode="r") as stream:
            yield cls.from_stream(stream, page_size=page_size, n_addresses=n_addresses, int_size=int_size)

    @staticmethod
    def compute_header_size(int_size: int) -> int:
        """Return the size (in bytes) of the header of an address file.

        Parameters
        ----------
        int_size : `int`
            Size of each integer in bytes.

        Returns
        -------
        size : `int`
            Size of the header in bytes.
        """
        return (
            1  # int_size
            + int_size  # number of rows
            + int_size  # number of addresses in each row
        )

    @staticmethod
    def compute_row_size(int_size: int, n_addresses: int) -> int:
        """Return the size (in bytes) of each row of an address file.

        Parameters
        ----------
        int_size : `int`
            Size of each integer in bytes.
        n_addresses : `int`
            Number of addresses in each row.

        Returns
        -------
        size : `int`
            Size of each row in bytes.
        """
        return (
            16  # uuid
            + int_size
            * (
                1  # index
                + 2 * n_addresses
            )
        )

    @property
    def row_size(self) -> int:
        """The size (in bytes) of each row of this address file."""
        return self.compute_row_size(self.int_size, self.n_addresses)

    def read_all(self) -> dict[uuid.UUID, AddressRow]:
        """Read all addresses in the file.

        Returns
        -------
        rows : `dict` [ `uuid.UUID`, `AddressRow` ]
            Mapping of loaded address rows, keyed by UUID.
        """
        # Skip any pages from the beginning that have already been read; this
        # nicely handles both the case where we already read everything (or
        # there was nothing to read) while giving us a page with a file offset
        # to start from.
        for page in self.pages:
            if not page.read:
                break
        else:
            return self.rows
        # Read the entire rest of the file into memory.
        self.stream.seek(page.file_offset)
        data = self.stream.read()
        buffer = BytesIO(data)
        # Shortcut out if we've already read everything, but don't bother
        # optimizing previous partial reads.
        while len(self.rows) < self.n_rows:
            self._read_row(buffer)
        # Delete all pages; they don't matter anymore, and that's easier than
        # updating them to reflect the reads we've done.
        self.pages.clear()
        return self.rows

    def find(self, key: uuid.UUID) -> AddressRow:
        """Read the row for the given UUID or integer index.

        Parameters
        ----------
        key : `uuid.UUID`
            UUID to find.

        Returns
        -------
        row : `AddressRow`
            Addresses for the given UUID.
        """
        if (row := self.rows.get(key)) is not None:
            return row
        if self.n_rows == 0 or not self.pages:
            raise LookupError(f"Address for {key} not found.")

        # Use a binary search to find the page containing the target UUID.
        left = 0
        right = len(self.pages) - 1
        while left <= right:
            mid = left + ((right - left) // 2)
            self._read_page(mid)
            if (row := self.rows.get(key)) is not None:
                return row
            bounds = self.page_bounds[mid]
            if key.int < bounds.uuid_int_begin:
                right = mid - 1
            elif key.int > bounds.uuid_int_end:
                left = mid + 1
            else:
                # Should have been on this page, but it wasn't.
                raise LookupError(f"Address for {key} not found.")

        # Ran out of pages to search.
        raise LookupError(f"Address for {key} not found.")

    def _read_page(self, page_index: int, page_stream: BytesIO | None = None) -> bool:
        page = self.pages[page_index]
        if page.read:
            return False
        if page_stream is None:
            self.stream.seek(page.file_offset)
            page_stream = BytesIO(self.stream.read(page.n_rows * self.row_size))
        row = self._read_row(page_stream)
        uuid_int_begin = row.key.int
        for _ in range(1, page.n_rows):
            row = self._read_row(page_stream)
        uuid_int_end = row.key.int + 1  # Python's loop scoping rules are actually useful here!
        page.read = True
        bounds = PageBounds(page_index=page_index, uuid_int_begin=uuid_int_begin, uuid_int_end=uuid_int_end)
        self.page_bounds[page_index] = bounds
        _LOG.debug("Read page %s with rows [%s:%s].", bounds, page.begin, page.end)
        return True

    def _read_row(self, page_stream: BytesIO) -> AddressRow:
        row = AddressRow.read(page_stream, self.n_addresses, self.int_size)
        self.rows[row.key] = row
        _LOG.debug("Read address row %s.", row)
        return row


@dataclasses.dataclass
class MultiblockWriter:
    """A helper object for writing multi-block files."""

    stream: IO[bytes]
    """A binary file-like object to write to."""

    int_size: int
    """Number of bytes to use for all integers."""

    file_size: int = 0
    """Running size of the full file."""

    addresses: dict[uuid.UUID, Address] = dataclasses.field(default_factory=dict)
    """Running map of all addresses added to the file so far.

    When the multi-block file is fully written, this is appended to the
    `AddressWriter.addresses` to write the corresponding address file.
    """

    @classmethod
    @contextmanager
    def open_in_zip(
        cls, zf: zipfile.ZipFile, name: str, int_size: int, use_tempfile: bool = False
    ) -> Iterator[MultiblockWriter]:
        """Open a writer for a file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to add the file to.
        name : `str`
            Base name for the multi-block file; an extension will be added.
        int_size : `int`
            Number of bytes to use for all integers.
        use_tempfile : `bool`, optional
            If `True`, send writes to a temporary file and only add the file to
            the zip archive when the context manager closes.  This involves
            more overall I/O, but it permits multiple multi-block files to be
            open for writing in the same zip archive at once.

        Returns
        -------
        writer : `contextlib.AbstractContextManager` [ `MultiblockWriter` ]
            Context manager that returns a writer when entered.
        """
        filename = f"{name}.mb"
        if use_tempfile:
            with tempfile.NamedTemporaryFile(suffix=filename) as tmp:
                yield MultiblockWriter(tmp, int_size)
                tmp.flush()
                zf.write(tmp.name, filename)
        else:
            with zf.open(f"{name}.mb", mode="w", force_zip64=True) as stream:
                yield MultiblockWriter(stream, int_size)

    def write_bytes(self, id: uuid.UUID, data: bytes) -> Address:
        """Write raw bytes to the multi-block file.

        Parameters
        ----------
        id : `uuid.UUID`
            Unique ID of the object described by this block.
        data : `bytes`
            Data to store directly.

        Returns
        -------
        address : `Address`
            Address of the bytes just written.
        """
        assert id not in self.addresses, "Duplicate write to multi-block file detected."
        self.stream.write(len(data).to_bytes(self.int_size))
        self.stream.write(data)
        block_size = len(data) + self.int_size
        address = Address(offset=self.file_size, size=block_size)
        self.file_size += block_size
        self.addresses[id] = address
        return address

    def write_model(self, id: uuid.UUID, model: pydantic.BaseModel, compressor: Compressor) -> Address:
        """Write raw bytes to the multi-block file.

        Parameters
        ----------
        id : `uuid.UUID`
            Unique ID of the object described by this block.
        model : `pydantic.BaseModel`
            Model to convert to JSON and compress.
        compressor : `Compressor`
            Object with a ``compress`` method that takes and returns `bytes`.

        Returns
        -------
        address : `Address`
            Address of the bytes just written.
        """
        json_data = model.model_dump_json().encode()
        compressed_data = compressor.compress(json_data)
        return self.write_bytes(id, compressed_data)


@dataclasses.dataclass
class MultiblockReader:
    """A helper object for reader multi-block files."""

    stream: IO[bytes]
    """A binary file-like object to read from."""

    int_size: int
    """Number of bytes to use for all integers."""

    @classmethod
    @contextmanager
    def open_in_zip(cls, zf: zipfile.ZipFile, name: str, *, int_size: int) -> Iterator[MultiblockReader]:
        """Open a reader for a file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to read the file from.
        name : `str`
            Base name for the multi-block file; an extension will be added.
        int_size : `int`
            Number of bytes to use for all integers.

        Returns
        -------
        reader : `contextlib.AbstractContextManager` [ `MultiblockReader` ]
            Context manager that returns a reader when entered.
        """
        with zf.open(f"{name}.mb", mode="r") as stream:
            yield MultiblockReader(stream, int_size)

    @classmethod
    def read_all_bytes_in_zip(
        cls, zf: zipfile.ZipFile, name: str, *, int_size: int, page_size: int
    ) -> Iterator[bytes]:
        """Iterate over all of the byte blocks in a file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to read the file from.
        name : `str`
            Base name for the multi-block file; an extension will be added.
        int_size : `int`
            Number of bytes to use for all integers.
        page_size : `int`
            Approximate number of bytes to read at a time.

        Returns
        -------
        byte_iter : `~collections.abc.Iterator` [ `bytes` ]
            Iterator over blocks.
        """
        with zf.open(f"{name}.mb", mode="r") as zf_stream:
            # The standard library typing of IO[bytes] tiers isn't consistent.
            buffered_stream = BufferedReader(zf_stream)  # type: ignore[type-var]
            size_data = buffered_stream.read(int_size)
            while size_data:
                internal_size = int.from_bytes(size_data)
                yield buffered_stream.read(internal_size)
                size_data = buffered_stream.read(int_size)

    @classmethod
    def read_all_models_in_zip(
        cls,
        zf: zipfile.ZipFile,
        name: str,
        model_type: type[_T],
        decompressor: Decompressor,
        *,
        int_size: int,
        page_size: int,
    ) -> Iterator[_T]:
        """Iterate over all of the models in a file in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to read the file from.
        name : `str`
            Base name for the multi-block file; an extension will be added.
        model_type : `type` [ `pydantic.BaseModel` ]
            Pydantic model to validate JSON with.
        decompressor : `Decompressor`
            Object with a ``decompress`` method that takes and returns `bytes`.
        int_size : `int`
            Number of bytes to use for all integers.
        page_size : `int`
            Approximate number of bytes to read at a time.

        Returns
        -------
        model_iter : `~collections.abc.Iterator` [ `pydantic.BaseModel` ]
            Iterator over model instances.
        """
        for compressed_data in cls.read_all_bytes_in_zip(zf, name, int_size=int_size, page_size=page_size):
            json_data = decompressor.decompress(compressed_data)
            yield model_type.model_validate_json(json_data)

    def read_bytes(self, address: Address) -> bytes | None:
        """Read raw bytes from the multi-block file.

        Parameters
        ----------
        address : `Address`
            Offset and size of the data to read.

        Returns
        -------
        data : `bytes` or `None`
            Data read directly, or `None` if the address has zero size.
        """
        if not address.size:
            return None
        self.stream.seek(address.offset)
        data = self.stream.read(address.size)
        internal_size = int.from_bytes(data[: self.int_size])
        data = data[self.int_size :]
        if len(data) != internal_size:
            raise InvalidQuantumGraphFileError(
                f"Internal size {internal_size} does not match loaded data size {len(data)}."
            )
        return data

    def read_model(self, address: Address, model_type: type[_T], decompressor: Decompressor) -> _T | None:
        """Read a single compressed JSON block.

        Parameters
        ----------
        address : `Address`
            Size and offset of the block.
        model_type : `type` [ `pydantic.BaseModel` ]
            Pydantic model to validate JSON with.
        decompressor : `Decompressor`
            Object with a ``decompress`` method that takes and returns `bytes`.

        Returns
        -------
        model : `pydantic.BaseModel`
            Validated model.
        """
        compressed_data = self.read_bytes(address)
        if compressed_data is None:
            return None
        json_data = decompressor.decompress(compressed_data)
        return model_type.model_validate_json(json_data)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "FORMAT_VERSION",
    "BaseQuantumGraph",
    "BaseQuantumGraphReader",
    "BipartiteEdgeInfo",
    "DatasetInfo",
    "HeaderModel",
    "QuantumInfo",
)
import dataclasses
import datetime
import getpass
import os
import sys
import uuid
import zipfile
from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    TypedDict,
)

import networkx
import networkx.algorithms.bipartite
import pydantic
import zstandard

from lsst.daf.butler import DataCoordinate, DataIdValue
from lsst.daf.butler._rubin import generate_uuidv7
from lsst.resources import ResourcePath, ResourcePathExpression

from ..pipeline_graph import DatasetTypeNode, Edge, PipelineGraph, TaskImportMode, TaskNode
from ..pipeline_graph.io import SerializedPipelineGraph
from ._multiblock import (
    DEFAULT_PAGE_SIZE,
    AddressReader,
    AddressWriter,
    Compressor,
    Decompressor,
)

if TYPE_CHECKING:
    from ..graph import QuantumGraph


# These aliases make it a lot easier how the various pydantic models are
# structured, but they're too verbose to be worth exporting to code outside the
# quantum_graph subpackage.
type TaskLabel = str
type DatasetTypeName = str
type ConnectionName = str
type DatasetIndex = int
type QuantumIndex = int
type DatastoreName = str
type DimensionElementName = str
type DataCoordinateValues = list[DataIdValue]


FORMAT_VERSION: int = 1
"""
File format version number for new files.

This applies to both predicted and provenance QGs, since they usually change
in concert.

CHANGELOG:

- 0: Initial version.
- 1: Switched from internal integer IDs to UUIDs in all models.
"""


class IncompleteQuantumGraphError(RuntimeError):
    pass


class HeaderModel(pydantic.BaseModel):
    """Data model for the header of a quantum graph file."""

    version: int = FORMAT_VERSION
    """File format / data model version number."""

    graph_type: str = ""
    """Type of quantum graph stored in this file."""

    inputs: list[str] = pydantic.Field(default_factory=list)
    """List of input collections used to build the quantum graph."""

    output: str | None = ""
    """Output CHAINED collection provided when building the quantum graph."""

    output_run: str = ""
    """Output RUN collection for all output datasets in this graph."""

    user: str = pydantic.Field(default_factory=getpass.getuser)
    """Username of the process that built this quantum graph."""

    timestamp: datetime.datetime = pydantic.Field(default_factory=datetime.datetime.now)
    """Timestamp for when this quantum graph was built.

    It is unspecified exactly which point during quantum-graph generation this
    timestamp is recorded.
    """

    command: str = pydantic.Field(default_factory=lambda: " ".join(sys.argv))
    """Command-line invocation that created this graph."""

    metadata: dict[str, Any] = pydantic.Field(default_factory=dict)
    """Free-form metadata associated with this quantum graph at build time."""

    int_size: int = 8
    """Number of bytes in the integers used in this file's multi-block and
    address files.
    """

    n_quanta: int = 0
    """Total number of quanta in this graph.

    This does not include special "init" quanta, but it does include quanta
    that were not loaded in a partial read (except when reading from an old
    quantum graph file).
    """

    n_datasets: int = 0
    """Total number of distinct datasets in the full graph.  This includes
    datasets whose related quanta were not loaded in a partial read (except
    when reading from an old quantum graph file).
    """

    n_task_quanta: dict[TaskLabel, int] = pydantic.Field(default_factory=dict)
    """Number of quanta for each task label.

    This does not include special "init" quanta, but it does include quanta
    that were not loaded in a partial read (except when reading from an old
    quantum graph file).
    """

    provenance_dataset_id: uuid.UUID = pydantic.Field(default_factory=generate_uuidv7)
    """The dataset ID for provenance quantum graph when it is ingested into
    a butler repository.
    """

    @classmethod
    def from_old_quantum_graph(cls, old_quantum_graph: QuantumGraph) -> HeaderModel:
        """Extract a header from an old `QuantumGraph` instance.

        Parameters
        ----------
        old_quantum_graph : `QuantumGraph`
            Quantum graph to extract a header from.

        Returns
        -------
        header : `PredictedHeaderModel`
            Header for a new predicted quantum graph.
        """
        metadata = dict(old_quantum_graph.metadata)
        metadata.pop("packages", None)
        if (time_str := metadata.pop("time", None)) is not None:
            timestamp = datetime.datetime.fromisoformat(time_str)
        else:
            timestamp = datetime.datetime.now()
        return cls(
            inputs=list(metadata.pop("input", []) or []),  # Guard against explicit None and missing key.
            output=metadata.pop("output", None),
            output_run=metadata.pop("output_run", ""),
            user=metadata.pop("user", ""),
            command=metadata.pop("full_command", ""),
            timestamp=timestamp,
            metadata=metadata,
        )

    def to_old_metadata(self) -> dict[str, Any]:
        """Return a dictionary using the key conventions used in old quantum
        graph files.
        """
        result = self.metadata.copy()
        result["input"] = self.inputs
        result["output"] = self.output
        result["output_run"] = self.output_run
        result["full_command"] = self.command
        result["user"] = self.user
        result["time"] = str(self.timestamp)
        return result

    # Work around the fact that Sphinx chokes on Pydantic docstring formatting,
    # when we inherit those docstrings in our public classes.
    if "sphinx" in sys.modules and not TYPE_CHECKING:

        def copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.copy`."""
            return super().copy(*args, **kwargs)

        def model_dump(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump`."""
            return super().model_dump(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_dump_json`."""
            return super().model_dump(*args, **kwargs)

        def model_copy(self, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_copy`."""
            return super().model_copy(*args, **kwargs)

        @classmethod
        def model_construct(cls, *args: Any, **kwargs: Any) -> Any:  # type: ignore[misc, override]
            """See `pydantic.BaseModel.model_construct`."""
            return super().model_construct(*args, **kwargs)

        @classmethod
        def model_json_schema(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_json_schema`."""
            return super().model_json_schema(*args, **kwargs)

        @classmethod
        def model_validate(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate`."""
            return super().model_validate(*args, **kwargs)

        @classmethod
        def model_validate_json(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_json`."""
            return super().model_validate_json(*args, **kwargs)

        @classmethod
        def model_validate_strings(cls, *args: Any, **kwargs: Any) -> Any:
            """See `pydantic.BaseModel.model_validate_strings`."""
            return super().model_validate_strings(*args, **kwargs)


class QuantumInfo(TypedDict):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    node data for a quantum.

    Since NetworkX types are not generic over their node mapping type, this has
    to be used explicitly, e.g.::

        node_data: QuantumInfo = xgraph.nodes[quantum_id]

    where ``xgraph`` can be either `BaseQuantumGraph.quantum_only_xgraph`
    or `BaseQuantumGraph.bipartite_xgraph`.
    """

    data_id: DataCoordinate
    """Data ID of the quantum."""

    task_label: str
    """Label of the task for this quantum."""

    pipeline_node: TaskNode
    """Node in the pipeline graph for this quantum's task."""


class DatasetInfo(TypedDict):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    node data for a dataset.

    Since NetworkX types are not generic over their node mapping type, this has
    to be used explicitly, e.g.::

        node_data: DatasetInfo = xgraph.nodes[dataset_id]

    where ``xgraph`` is from the `BaseQuantumGraph.bipartite_xgraph` property.
    """

    data_id: DataCoordinate
    """Data ID of the dataset."""

    dataset_type_name: DatasetTypeName
    """Name of the type of this dataset.

    This is always the general dataset type that matches the data repository
    storage class, which may differ from any particular task-adapted dataset
    type whose storage class has been overridden to match the task connections.
    This means is it never a component.
    """

    run: str
    """Name of the `~lsst.daf.butler.CollectionType.RUN` collection that holds
    or will hold this dataset.
    """

    pipeline_node: DatasetTypeNode
    """Node in the pipeline graph for this dataset's type."""


class BipartiteEdgeInfo(TypedDict):
    """A typed dictionary that annotates the attributes of the NetworkX graph
    edge data in a bipartite graph.
    """

    is_read: bool
    """`True` if this is a dataset -> quantum edge; `False` if it is a
    quantum -> dataset edge.
    """

    pipeline_edges: list[Edge]
    """Corresponding edges in the pipeline graph.

    Note that there may be more than one pipeline edge since a quantum can
    consume a particular dataset via multiple connections.
    """


class BaseQuantumGraph(ABC):
    """An abstract base for quantum graphs.

    Parameters
    ----------
    header : `HeaderModel`
        Structured metadata for the graph.
    pipeline_graph : `.pipeline_graph.PipelineGraph`
        Graph of tasks and dataset types.  May contain a superset of the tasks
        and dataset types that actually have quanta and datasets in the quantum
        graph.
    """

    def __init__(self, header: HeaderModel, pipeline_graph: PipelineGraph):
        self.header = header
        self.pipeline_graph = pipeline_graph

    @property
    @abstractmethod
    def quanta_by_task(self) -> Mapping[str, Mapping[DataCoordinate, uuid.UUID]]:
        """A nested mapping of all quanta, keyed first by task name and then by
        data ID.

        Notes
        -----
        Partial loads may not fully populate this mapping, but it can always
        be accessed.

        The returned object may be an internal dictionary; as the type
        annotation indicates, it should not be modified in place.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def datasets_by_type(self) -> Mapping[str, Mapping[DataCoordinate, uuid.UUID]]:
        """A nested mapping of all datasets, keyed first by dataset type name
        and then by data ID.

        Notes
        -----
        Partial loads may not fully populate this mapping, but it can always
        be accessed.

        The returned object may be an internal dictionary; as the type
        annotation indicates, it should not be modified in place.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def quantum_only_xgraph(self) -> networkx.DiGraph:
        """A directed acyclic graph with quanta as nodes and datasets elided.

        Notes
        -----
        Partial loads may not fully populate this graph, but it can always be
        accessed.

        Node state dictionaries are described by the `QuantumInfo` type
        (or a subtype thereof).

        The returned object is a read-only view of an internal one.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def bipartite_xgraph(self) -> networkx.DiGraph:
        """A directed acyclic graph with quantum and dataset nodes.

        Notes
        -----
        Partial loads may not fully populate this graph, but it can always be
        accessed.

        Node state dictionaries are described by the `QuantumInfo` and
        `DatasetInfo` types (or a subtypes thereof).  Edges have state
        dictionaries described by `BipartiteEdgeInfo`.

        The returned object is a read-only view of an internal one.
        """
        raise NotImplementedError()


@dataclasses.dataclass
class BaseQuantumGraphWriter:
    """A helper class for writing quantum graphs."""

    zf: zipfile.ZipFile
    """The zip archive that represents the quantum graph on disk."""

    compressor: Compressor
    """A compressor for all compressed JSON blocks."""

    address_writer: AddressWriter
    """A helper object for reading addresses into the multi-block files."""

    int_size: int
    """Size (in bytes) used to write integers to binary files."""

    @classmethod
    @contextmanager
    def open(
        cls,
        uri: ResourcePathExpression,
        header: HeaderModel,
        pipeline_graph: PipelineGraph,
        *,
        address_filename: str,
        cdict_data: bytes | None = None,
        zstd_level: int = 10,
    ) -> Iterator[Self]:
        uri = ResourcePath(uri, forceDirectory=False)
        address_writer = AddressWriter()
        if uri.isLocal:
            os.makedirs(uri.dirname().ospath, exist_ok=True)
        cdict = zstandard.ZstdCompressionDict(cdict_data) if cdict_data is not None else None
        compressor = zstandard.ZstdCompressor(level=zstd_level, dict_data=cdict)
        with uri.open(mode="wb") as stream:
            with zipfile.ZipFile(stream, mode="w", compression=zipfile.ZIP_STORED) as zf:
                self = cls(zf, compressor, address_writer, header.int_size)
                self.write_single_model("header", header)
                if cdict_data is not None:
                    zf.writestr("compression_dict", cdict_data)
                self.write_single_model("pipeline_graph", SerializedPipelineGraph.serialize(pipeline_graph))
                yield self
                address_writer.write_to_zip(zf, address_filename, int_size=self.int_size)

    def write_single_model(self, name: str, model: pydantic.BaseModel) -> None:
        """Write a single compressed JSON block as a 'file' in a zip archive.

        Parameters
        ----------
        name : `str`
            Base name of the file.  An extension will be added.
        model : `pydantic.BaseModel`
            Pydantic model to convert to JSON.
        """
        json_data = model.model_dump_json().encode()
        self.write_single_block(name, json_data)

    def write_single_block(self, name: str, json_data: bytes) -> None:
        """Write a single compressed JSON block as a 'file' in a zip archive.

        Parameters
        ----------
        name : `str`
            Base name of the file.  An extension will be added.
        json_data : `bytes`
            Raw JSON to compress and write.
        """
        json_data = self.compressor.compress(json_data)
        self.zf.writestr(f"{name}.json.zst", json_data)


@dataclasses.dataclass
class BaseQuantumGraphReader:
    """A helper class for reading quantum graphs."""

    header: HeaderModel
    """Header metadata for the quantum graph."""

    pipeline_graph: PipelineGraph
    """Graph of tasks and dataset type names that appear in the quantum
    graph.
    """

    zf: zipfile.ZipFile
    """The zip archive that represents the quantum graph on disk."""

    decompressor: Decompressor
    """A decompressor for all compressed JSON blocks."""

    address_reader: AddressReader
    """A helper object for reading addresses into the multi-block files."""

    page_size: int
    """Approximate number of bytes to read at a time.

    Note that this does not set a page size for *all* reads, but it
    does affect the smallest, most numerous reads.
    """

    @classmethod
    @contextmanager
    def _open(
        cls,
        uri: ResourcePathExpression,
        *,
        address_filename: str,
        graph_type: str,
        n_addresses: int,
        page_size: int | None = None,
        import_mode: TaskImportMode = TaskImportMode.ASSUME_CONSISTENT_EDGES,
    ) -> Iterator[Self]:
        """Construct a reader from a URI.

        Parameters
        ----------
        uri : convertible to `lsst.resources.ResourcePath`
            URI to open.  Should have a ``.qg`` extension.
        address_filename : `str`
            Base filename for the address file.
        graph_type : `str`
            Value to expect for `HeaderModel.graph_type`.
        n_addresses : `int`
            Number of addresses to expect per row in the address file.
        page_size : `int`, optional
            Approximate number of bytes to read at once from address files.
            Note that this does not set a page size for *all* reads, but it
            does affect the smallest, most numerous reads.  When `None`, the
            ``LSST_QG_PAGE_SIZE`` environment variable is checked before
            falling back to a default of 5MB.
        import_mode : `..pipeline_graph.TaskImportMode`, optional
            How to handle importing the task classes referenced in the pipeline
            graph.

        Returns
        -------
        reader : `contextlib.AbstractContextManager` [ \
                `PredictedQuantumGraphReader` ]
            A context manager that returns the reader when entered.
        """
        if page_size is None:
            page_size = int(os.environ.get("LSST_QG_PAGE_SIZE", DEFAULT_PAGE_SIZE))
        uri = ResourcePath(uri)
        cdict: zstandard.ZstdCompressionDict | None = None
        with uri.open(mode="rb") as zf_stream:
            with zipfile.ZipFile(zf_stream, "r") as zf:
                if (cdict_path := zipfile.Path(zf, "compression_dict")).exists():
                    cdict = zstandard.ZstdCompressionDict(cdict_path.read_bytes())
                decompressor = zstandard.ZstdDecompressor(cdict)
                header = cls._read_single_block_static("header", HeaderModel, zf, decompressor)
                if not header.graph_type == graph_type:
                    raise TypeError(f"Header is for a {header.graph_type!r} graph, not {graph_type!r} graph.")
                serialized_pipeline_graph = cls._read_single_block_static(
                    "pipeline_graph", SerializedPipelineGraph, zf, decompressor
                )
                pipeline_graph = serialized_pipeline_graph.deserialize(import_mode)
                with AddressReader.open_in_zip(
                    zf,
                    address_filename,
                    page_size=page_size,
                    int_size=header.int_size,
                    n_addresses=n_addresses,
                ) as address_reader:
                    yield cls(
                        header=header,
                        pipeline_graph=pipeline_graph,
                        zf=zf,
                        decompressor=decompressor,
                        address_reader=address_reader,
                        page_size=page_size,
                    )

    @staticmethod
    def _read_single_block_static[T: pydantic.BaseModel](
        name: str, model_type: type[T], zf: zipfile.ZipFile, decompressor: Decompressor
    ) -> T:
        """Read a single compressed JSON block from a 'file' in a zip archive.

        Parameters
        ----------
        zf : `zipfile.ZipFile`
            Zip archive to read the file from.
        name : `str`
            Base name of the file.  An extension will be added.
        model_type : `type` [ `pydantic.BaseModel` ]
            Pydantic model to validate JSON with.
        decompressor : `Decompressor`
            Object with a `decompress` method that takes and returns `bytes`.

        Returns
        -------
        model : `pydantic.BaseModel`
            Validated model.
        """
        compressed_data = zf.read(f"{name}.json.zst")
        json_data = decompressor.decompress(compressed_data)
        return model_type.model_validate_json(json_data)

    def _read_single_block[T: pydantic.BaseModel](self, name: str, model_type: type[T]) -> T:
        """Read a single compressed JSON block from a 'file' in a zip archive.

        Parameters
        ----------
        name : `str`
            Base name of the file.  An extension will be added.
        model_type : `type` [ `pydantic.BaseModel` ]
            Pydantic model to validate JSON with.

        Returns
        -------
        model : `pydantic.BaseModel`
            Validated model.
        """
        return self._read_single_block_static(name, model_type, self.zf, self.decompressor)

    def _read_single_block_raw(self, name: str) -> bytes:
        """Read a single compressed block from a 'file' in a zip archive.

        Parameters
        ----------
        name : `str`
            Base name of the file.  An extension will be added.

        Returns
        -------
        data : `bytes`
            Decompressed bytes.
        """
        compressed_data = self.zf.read(f"{name}.json.zst")
        return self.decompressor.decompress(compressed_data)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import os
import shutil
import tempfile
import unittest

import lsst.pipe.base.quantum_provenance_graph as qpg
import lsst.utils.tests
from lsst.daf.butler import Butler, DatasetType
from lsst.pipe.base import PipelineGraph, QuantumSuccessCaveats, RepeatableQuantumError
from lsst.pipe.base.simple_pipeline_executor import SimplePipelineExecutor
from lsst.pipe.base.tests.mocks import (
    DirectButlerRepo,
    DynamicConnectionConfig,
    DynamicTestPipelineTask,
    DynamicTestPipelineTaskConfig,
    MockStorageClass,
    get_mock_name,
)

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class SimplePipelineExecutorTests(lsst.utils.tests.TestCase):
    """Test the SimplePipelineExecutor API.

    Because SimplePipelineExecutor is the easiest way to run simple pipelines
    in tests, this has also become a home for tests of execution edge cases
    that don't have a clear home in other test files.
    """

    def setUp(self):
        self.path = tempfile.mkdtemp()
        # standalone parameter forces the returned config to also include
        # the information from the search paths.
        config = Butler.makeRepo(self.path, standalone=True, searchPaths=[os.path.join(TESTDIR, "config")])
        self.butler = SimplePipelineExecutor.prep_butler(config, [], "fake")
        self.enterContext(self.butler)
        self.butler.registry.registerDatasetType(
            DatasetType(
                "input",
                dimensions=self.butler.dimensions.empty,
                storageClass="StructuredDataDict",
            )
        )
        self.butler.put({"zero": 0}, "input")
        MockStorageClass.get_or_register_mock("StructuredDataDict")
        MockStorageClass.get_or_register_mock("TaskMetadataLike")

    def tearDown(self):
        shutil.rmtree(self.path, ignore_errors=True)

    def test_from_task_class(self):
        """Test executing a single quantum with an executor created by the
        `from_task_class` factory method, and the
        `SimplePipelineExecutor.as_generator` method.
        """
        config = DynamicTestPipelineTaskConfig()
        config.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        executor = SimplePipelineExecutor.from_task_class(
            DynamicTestPipelineTask,
            config=config,
            butler=self.butler,
            label="a",
        )
        (quantum,) = executor.as_generator(register_dataset_types=True)
        self.assertEqual(self.butler.get("output").storage_class, get_mock_name("StructuredDataDict"))

    def test_metadata_input(self):
        """Test two tasks where the output uses metadata from input."""
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_b = DynamicTestPipelineTaskConfig()
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        config_b.inputs["in_metadata"] = DynamicConnectionConfig(
            dataset_type_name="a_metadata", storage_class="TaskMetadata"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        pipeline_graph.add_task("b", DynamicTestPipelineTask, config_b)
        executor = SimplePipelineExecutor.from_pipeline_graph(pipeline_graph, butler=self.butler)
        quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertEqual(len(quanta), 2)
        output = self.butler.get("output")
        self.assertEqual(output.quantum.inputs["in_metadata"][0].original_type, "lsst.pipe.base.TaskMetadata")

    def test_optional_intermediate(self):
        """Test a pipeline task with an optional regular input that is produced
        by another task.
        """
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.fail_exception = "lsst.pipe.base.NoWorkFound"
        config_a.fail_condition = "1=1"  # butler query expression that is true
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        config_b = DynamicTestPipelineTaskConfig()
        config_b.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict", minimum=0
        )
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        pipeline_graph.add_task("b", DynamicTestPipelineTask, config_b)
        executor = SimplePipelineExecutor.from_pipeline_graph(pipeline_graph, butler=self.butler)
        quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertEqual(len(quanta), 2)
        # Both quanta ran successfully (NoWorkFound is a success).
        self.assertTrue(self.butler.exists("a_metadata"))
        self.assertTrue(self.butler.exists("b_metadata"))
        # The intermediate dataset was not written, but the final output was.
        self.assertFalse(self.butler.exists("intermediate"))
        self.assertTrue(self.butler.exists("output"))

    def test_optional_input(self):
        """Test a pipeline task with an optional regular input that is an
        overall input to the pipeline.
        """
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i1"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.outputs["i2"] = DynamicConnectionConfig(
            dataset_type_name="input_2",
            storage_class="StructuredDataDict",  # will never exist
        )
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        executor = SimplePipelineExecutor.from_pipeline_graph(pipeline_graph, butler=self.butler)
        quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertEqual(len(quanta), 1)
        # The quanta ran successfully.
        self.assertTrue(self.butler.exists("a_metadata"))
        # The final output was written.
        self.assertTrue(self.butler.exists("output"))

    def test_from_pipeline_file(self) -> None:
        """Test executing a two quanta from different configurations of the
        same task, with an executor created by the `from_pipeline_filename`
        factory method, and the `SimplePipelineExecutor.run` method.
        """
        filename = os.path.join(TESTDIR, "pipelines", "pipeline_simple.yaml")
        executor = SimplePipelineExecutor.from_pipeline_filename(filename, butler=self.butler)
        self._test_pipeline_file(executor)

    def test_use_local_butler(self) -> None:
        """Test generating a local butler repository from a pipeline, then
        running that pipeline using the local butler.
        """
        # Test a trivial pipeline that has only dataset types with empty
        # dimensions.
        filename = os.path.join(TESTDIR, "pipelines", "pipeline_simple.yaml")
        executor = SimplePipelineExecutor.from_pipeline_filename(
            filename, butler=self.butler, output="u/someone/pipeline"
        )
        with tempfile.TemporaryDirectory() as tempdir:
            root = os.path.join(tempdir, "butler_root")
            local_butler = executor.use_local_butler(root)
            self.enterContext(local_butler)
            self._test_pipeline_file(executor)

        # Test a more complicated pipeline involving dataset types with
        # non-empty dimensions.  This will require dimension records to be
        # copied into the destination repository.
        with (
            tempfile.TemporaryDirectory() as tempdir,
            DirectButlerRepo.make_temporary("base.yaml", "spatial.yaml") as (helper, root),
        ):
            helper.add_task(dimensions=["visit", "detector"])
            qg = helper.make_quantum_graph(output="out_chain")
            executor = SimplePipelineExecutor(qg, helper.butler)
            output_butler_root = os.path.join(tempdir, "root")
            local_butler = executor.use_local_butler(output_butler_root)
            self.enterContext(local_butler)
            executor.run(register_dataset_types=True)
            output_butler = Butler.from_config(output_butler_root)
            self.enterContext(output_butler)
            ref = output_butler.find_dataset(
                "dataset_auto1",
                collections="out_chain",
                dimension_records=True,
                instrument="Cam1",
                detector=1,
                visit=1,
            )
            self.assertIsNotNone(ref)
            assert ref is not None  # For linters.
            self.assertIsNotNone(output_butler.get(ref))
            # Check that dimension records are present in the output Butler.
            self.assertEqual(ref.dataId.records["visit"].science_program, "test_survey")
            self.assertTrue(ref.dataId.records["visit_detector_region"].region.contains(0.004, 0.020))

    def _test_pipeline_file(self, executor: SimplePipelineExecutor) -> None:
        quanta = executor.run(register_dataset_types=True, save_versions=False)
        self.assertEqual(len(quanta), 2)
        self.assertEqual(
            executor.butler.get("intermediate").storage_class, get_mock_name("StructuredDataDict")
        )
        self.assertEqual(executor.butler.get("output").storage_class, get_mock_name("StructuredDataDict"))

    def test_partial_outputs_success(self):
        """Test executing two quanta where the first raises
        `lsst.pipe.base.AnnotatedPartialOutputsError` and its output is an
        optional input to the second, while configuring the executor to
        consider this a success.
        """
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.fail_exception = "lsst.pipe.base.AnnotatedPartialOutputsError"
        config_a.fail_condition = "1=1"  # butler query expression that is true
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        config_b = DynamicTestPipelineTaskConfig()
        config_b.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict", minimum=0
        )
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        pipeline_graph.add_task("b", DynamicTestPipelineTask, config_b)
        # Consider the partial a success and proceed.
        executor = SimplePipelineExecutor.from_pipeline_graph(
            pipeline_graph, butler=self.butler, raise_on_partial_outputs=False
        )
        (_, _) = executor.as_generator(register_dataset_types=True)
        self.assertFalse(self.butler.exists("intermediate"))
        self.assertEqual(self.butler.get("output").storage_class, get_mock_name("StructuredDataDict"))
        prov = qpg.QuantumProvenanceGraph(self.butler, [executor.quantum_graph], read_caveats="exhaustive")
        (quantum_key_a,) = prov.quanta["a"]
        quantum_info_a = prov.get_quantum_info(quantum_key_a)
        _, quantum_run_a = qpg.QuantumRun.find_final(quantum_info_a)
        self.assertEqual(
            quantum_run_a.caveats,
            QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
            | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
            | QuantumSuccessCaveats.PARTIAL_OUTPUTS_ERROR,
        )
        self.assertEqual(
            quantum_run_a.exception.type_name,
            "lsst.pipe.base.tests.mocks.MockAlgorithmError",
        )
        self.assertEqual(
            quantum_run_a.exception.metadata,
            {"badness": 12},
        )
        (quantum_key_b,) = prov.quanta["b"]
        quantum_info_b = prov.get_quantum_info(quantum_key_b)
        _, quantum_run_b = qpg.QuantumRun.find_final(quantum_info_b)
        self.assertEqual(quantum_run_b.caveats, QuantumSuccessCaveats.NO_CAVEATS)
        prov_summary = prov.to_summary(self.butler)
        # One partial-outputs case, with an empty data ID:
        self.assertEqual(prov_summary.tasks["a"].caveats, {"*P": [{}]})
        self.assertEqual(
            prov_summary.tasks["a"].exceptions.keys(), {"lsst.pipe.base.tests.mocks.MockAlgorithmError"}
        )
        self.assertEqual(
            prov_summary.tasks["a"]
            .exceptions["lsst.pipe.base.tests.mocks.MockAlgorithmError"][0]
            .exception.metadata,
            {"badness": 12},
        )
        # No caveats for the second task, since it didn't need the first task's
        # output anyway.
        self.assertEqual(prov_summary.tasks["b"].caveats, {})
        self.assertEqual(prov_summary.tasks["b"].exceptions, {})
        # Check table forms for summaries of the same information.
        quantum_table = prov_summary.make_quantum_table()
        self.assertEqual(list(quantum_table["Task"]), ["a", "b"])
        self.assertEqual(list(quantum_table["Unknown"]), [0, 0])
        self.assertEqual(list(quantum_table["Successful"]), [1, 1])
        self.assertEqual(list(quantum_table["Caveats"]), ["*P(1)", ""])
        self.assertEqual(list(quantum_table["Blocked"]), [0, 0])
        self.assertEqual(list(quantum_table["Failed"]), [0, 0])
        self.assertEqual(list(quantum_table["Wonky"]), [0, 0])
        self.assertEqual(list(quantum_table["TOTAL"]), [1, 1])
        self.assertEqual(list(quantum_table["EXPECTED"]), [1, 1])
        dataset_table = prov_summary.make_dataset_table()
        self.assertEqual(
            list(dataset_table["Dataset"]),
            ["intermediate", "a_metadata", "a_log", "output", "b_metadata", "b_log"],
        )
        self.assertEqual(list(dataset_table["Visible"]), [0, 1, 1, 1, 1, 1])
        self.assertEqual(list(dataset_table["Shadowed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Predicted Only"]), [1, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Unsuccessful"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Cursed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["TOTAL"]), [1, 1, 1, 1, 1, 1])
        self.assertEqual(list(dataset_table["EXPECTED"]), [1, 1, 1, 1, 1, 1])
        exception_table = prov_summary.make_exception_table()
        self.assertEqual(list(exception_table["Task"]), ["a"])
        self.assertEqual(
            list(exception_table["Exception"]), ["lsst.pipe.base.tests.mocks.MockAlgorithmError"]
        )
        self.assertEqual(list(exception_table["Count"]), [1])
        bad_quantum_tables = prov_summary.make_bad_quantum_tables()
        self.assertEqual(bad_quantum_tables.keys(), {"a"})
        self.assertEqual(list(bad_quantum_tables["a"]["Status(Caveats)"]), ["SUCCESSFUL(P)"])
        self.assertEqual(list(bad_quantum_tables["a"]["Exception"]), ["MockAlgorithmError"])
        self.assertFalse(prov_summary.make_bad_dataset_tables())

    def test_no_work_found(self):
        """Test executing two quanta where the first raises
        `NoWorkFound` in `runQuantum`, leading the next to raise `NoWorkFound`
        in `adjustQuantum`.
        """
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.fail_exception = "lsst.pipe.base.NoWorkFound"
        config_a.fail_condition = "1=1"  # butler query expression that is true
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        config_b = DynamicTestPipelineTaskConfig()
        config_b.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        pipeline_graph.add_task("b", DynamicTestPipelineTask, config_b)
        # Consider the partial a success and proceed.
        executor = SimplePipelineExecutor.from_pipeline_graph(
            pipeline_graph, butler=self.butler, raise_on_partial_outputs=False
        )
        (_, _) = executor.as_generator(register_dataset_types=True)
        prov = qpg.QuantumProvenanceGraph()
        prov.assemble_quantum_provenance_graph(self.butler, [executor.quantum_graph])
        (quantum_key_a,) = prov.quanta["a"]
        quantum_info_a = prov.get_quantum_info(quantum_key_a)
        _, quantum_run_a = qpg.QuantumRun.find_final(quantum_info_a)
        self.assertEqual(
            quantum_run_a.caveats,
            QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
            | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
            | QuantumSuccessCaveats.NO_WORK,
        )
        (quantum_key_b,) = prov.quanta["b"]
        quantum_info_b = prov.get_quantum_info(quantum_key_b)
        _, quantum_run_b = qpg.QuantumRun.find_final(quantum_info_b)
        self.assertEqual(
            quantum_run_b.caveats,
            QuantumSuccessCaveats.ALL_OUTPUTS_MISSING
            | QuantumSuccessCaveats.ANY_OUTPUTS_MISSING
            | QuantumSuccessCaveats.ADJUST_QUANTUM_RAISED
            | QuantumSuccessCaveats.NO_WORK,
        )
        prov_summary = prov.to_summary(self.butler)
        # One NoWorkFound, raised by runQuantum, with an empty data ID:
        self.assertEqual(prov_summary.tasks["a"].caveats, {"*N": [{}]})
        self.assertEqual(prov_summary.tasks["a"].exceptions, {})
        # One NoWorkFound, raised by adjustQuantum, with an empty data ID.
        self.assertEqual(prov_summary.tasks["b"].caveats, {"*A": [{}]})
        self.assertEqual(prov_summary.tasks["b"].exceptions, {})
        # Check table forms for summaries of the same information.
        quantum_table = prov_summary.make_quantum_table()
        self.assertEqual(list(quantum_table["Task"]), ["a", "b"])
        self.assertEqual(list(quantum_table["Unknown"]), [0, 0])
        self.assertEqual(list(quantum_table["Successful"]), [1, 1])
        self.assertEqual(list(quantum_table["Caveats"]), ["*N(1)", "*A(1)"])
        self.assertEqual(list(quantum_table["Blocked"]), [0, 0])
        self.assertEqual(list(quantum_table["Failed"]), [0, 0])
        self.assertEqual(list(quantum_table["Wonky"]), [0, 0])
        self.assertEqual(list(quantum_table["TOTAL"]), [1, 1])
        self.assertEqual(list(quantum_table["EXPECTED"]), [1, 1])
        dataset_table = prov_summary.make_dataset_table()
        self.assertEqual(
            list(dataset_table["Dataset"]),
            ["intermediate", "a_metadata", "a_log", "output", "b_metadata", "b_log"],
        )
        self.assertEqual(list(dataset_table["Visible"]), [0, 1, 1, 0, 1, 1])
        self.assertEqual(list(dataset_table["Shadowed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Predicted Only"]), [1, 0, 0, 1, 0, 0])
        self.assertEqual(list(dataset_table["Unsuccessful"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Cursed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["TOTAL"]), [1, 1, 1, 1, 1, 1])
        self.assertEqual(list(dataset_table["EXPECTED"]), [1, 1, 1, 1, 1, 1])
        self.assertFalse(prov_summary.make_exception_table())
        self.assertFalse(prov_summary.make_bad_quantum_tables())
        self.assertFalse(prov_summary.make_bad_dataset_tables())

    def test_partial_outputs_failure(self):
        """Test executing two quanta where the first raises
        `lsst.pipe.base.AnnotatedPartialOutputsError` and its output is an
        optional input to the second, while configuring the executor to
        consider this a failure.
        """
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.fail_exception = "lsst.pipe.base.AnnotatedPartialOutputsError"
        config_a.fail_condition = "1=1"  # butler query expression that is true
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        config_b = DynamicTestPipelineTaskConfig()
        config_b.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict", minimum=0
        )
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        pipeline_graph = PipelineGraph()
        pipeline_graph.add_task("a", DynamicTestPipelineTask, config_a)
        pipeline_graph.add_task("b", DynamicTestPipelineTask, config_b)
        executor = SimplePipelineExecutor.from_pipeline_graph(
            pipeline_graph,
            butler=self.butler,
            raise_on_partial_outputs=True,
        )
        # The executor should raise the chained exception
        # (RepeatableQuantumError, since that's what the mocking system in
        # pipe_base uses here), not AnnotatedPartialOutputsError.
        with self.assertRaises(RepeatableQuantumError):
            executor.run(register_dataset_types=True)
        self.assertFalse(self.butler.exists("intermediate"))
        self.assertFalse(self.butler.exists("output"))
        prov = qpg.QuantumProvenanceGraph()
        prov.assemble_quantum_provenance_graph(self.butler, [executor.quantum_graph])
        (quantum_key_a,) = prov.quanta["a"]
        quantum_info_a = prov.get_quantum_info(quantum_key_a)
        _, quantum_run_a = qpg.QuantumRun.find_final(quantum_info_a)
        self.assertEqual(quantum_run_a.status, qpg.QuantumRunStatus.FAILED)
        self.assertIsNone(quantum_run_a.caveats)
        self.assertIsNone(quantum_run_a.exception)
        (quantum_key_b,) = prov.quanta["b"]
        quantum_info_b = prov.get_quantum_info(quantum_key_b)
        self.assertEqual(quantum_info_b["status"], qpg.QuantumInfoStatus.BLOCKED)
        prov_summary = prov.to_summary(self.butler)
        # One partial-outputs failure case for the first task.
        self.assertEqual(prov_summary.tasks["a"].n_failed, 1)
        # No direct failures, but one blocked for the second
        self.assertEqual(prov_summary.tasks["b"].n_failed, 0)
        self.assertEqual(prov_summary.tasks["b"].n_blocked, 1)
        # Check table forms for summaries of the same information.
        quantum_table = prov_summary.make_quantum_table()
        self.assertEqual(list(quantum_table["Task"]), ["a", "b"])
        self.assertEqual(list(quantum_table["Unknown"]), [0, 0])
        self.assertEqual(list(quantum_table["Successful"]), [0, 0])
        self.assertEqual(list(quantum_table["Caveats"]), ["", ""])
        self.assertEqual(list(quantum_table["Blocked"]), [0, 1])
        self.assertEqual(list(quantum_table["Failed"]), [1, 0])
        self.assertEqual(list(quantum_table["Wonky"]), [0, 0])
        self.assertEqual(list(quantum_table["TOTAL"]), [1, 1])
        self.assertEqual(list(quantum_table["EXPECTED"]), [1, 1])
        dataset_table = prov_summary.make_dataset_table()
        self.assertEqual(
            list(dataset_table["Dataset"]),
            ["intermediate", "a_metadata", "a_log", "output", "b_metadata", "b_log"],
        )
        # Note that a_log is UNSUCCESSFUL, not VISIBLE, despite being present
        # in butler because those categories are mutually exclusive and we
        # don't want to consider any outputs of failed quanta to be successful.
        self.assertEqual(list(dataset_table["Visible"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Shadowed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Predicted Only"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["Unsuccessful"]), [1, 1, 1, 1, 1, 1])
        self.assertEqual(list(dataset_table["Cursed"]), [0, 0, 0, 0, 0, 0])
        self.assertEqual(list(dataset_table["TOTAL"]), [1, 1, 1, 1, 1, 1])
        self.assertEqual(list(dataset_table["EXPECTED"]), [1, 1, 1, 1, 1, 1])
        self.assertFalse(prov_summary.make_exception_table())
        bad_quantum_tables = prov_summary.make_bad_quantum_tables()
        self.assertEqual(bad_quantum_tables.keys(), {"a"})
        self.assertEqual(list(bad_quantum_tables["a"]["Status(Caveats)"]), ["FAILED"])
        self.assertEqual(list(bad_quantum_tables["a"]["Exception"]), [""])
        self.assertFalse(prov_summary.make_bad_dataset_tables())

    def test_existence_check_skips(self):
        """Test that pre-execution existence checks are not performed for
        overall-input datasets, as this those checks could otherwise mask
        repository configuration problems or downtime as NoWorkFound cases.
        """
        # First we configure and execute task A, which is just a way to get a
        # MockDataset in the repo for us to play with; the important test can't
        # use the non-mock 'input' dataset because the mock runQuantum only
        # actually reads MockDatasets.
        config_a = DynamicTestPipelineTaskConfig()
        config_a.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="input", storage_class="StructuredDataDict", mock_storage_class=False
        )
        config_a.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict"
        )
        executor_a = SimplePipelineExecutor.from_task_class(
            DynamicTestPipelineTask,
            config=config_a,
            butler=self.butler,
            label="a",
        )
        executor_a.run(register_dataset_types=True)
        # Now we can do the real test.
        config_b = DynamicTestPipelineTaskConfig()
        config_b.inputs["i"] = DynamicConnectionConfig(
            dataset_type_name="intermediate", storage_class="StructuredDataDict", minimum=0
        )
        config_b.outputs["o"] = DynamicConnectionConfig(
            dataset_type_name="output", storage_class="StructuredDataDict"
        )
        butler = self.butler.clone(run="new_run")
        executor_b = SimplePipelineExecutor.from_task_class(
            DynamicTestPipelineTask,
            config=config_b,
            butler=butler,
            label="b",
            attach_datastore_records=True,
        )
        # Delete the input dataset after the QG has already been built.
        intermediate_refs = butler.query_datasets("intermediate")
        self.assertEqual(len(intermediate_refs), 1)
        butler.pruneDatasets(intermediate_refs, purge=True, unstore=True)
        with self.assertRaises(FileNotFoundError):
            # We should get an exception rather than NoWorkFound, because for
            # this single-task pipeline, the missing dataset is an
            # overall-input (name notwithstanding).
            executor_b.run(register_dataset_types=True)


class MemoryTester(lsst.utils.tests.MemoryTestCase):
    """Generic tests for file leaks."""


def setup_module(module):
    """Set up the module for pytest.

    Parameters
    ----------
    module : `~types.ModuleType`
        Module to set up.
    """
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()
